import { Test, TestingModule } from '@nestjs/testing';
import {
  AggregateID,
  NotFoundException,
  UniqueConstraintException,
} from '@mobicoop/ddd-library';
import { USER_REPOSITORY } from '@modules/user/user.di-tokens';
import { UserEntity } from '@modules/user/core/domain/user.entity';
import {
  EmailAlreadyExistsException,
  PhoneAlreadyExistsException,
} from '@modules/user/core/domain/user.errors';
import { UpdateUserRequestDto } from '@modules/user/interface/grpc-controllers/dtos/update-user.request.dto';
import { UpdateUserService } from '@modules/user/core/application/commands/update-user/update-user.service';
import { UpdateUserCommand } from '@modules/user/core/application/commands/update-user/update-user.command';

const updateFirstNameUserRequest: UpdateUserRequestDto = {
  id: 'c97b1783-76cf-4840-b298-b90b13c58894',
  firstName: 'Johnny',
};

const updateEmailUserRequest: UpdateUserRequestDto = {
  id: 'c97b1783-76cf-4840-b298-b90b13c58894',
  email: 'john.doe@already.exists.email.com',
};

const updatePhoneUserRequest: UpdateUserRequestDto = {
  id: 'c97b1783-76cf-4840-b298-b90b13c58894',
  phone: '+33611223344',
};

const now = new Date();
const userToUpdate: UserEntity = new UserEntity({
  id: 'c97b1783-76cf-4840-b298-b90b13c58894',
  createdAt: now,
  updatedAt: now,
  props: {
    firstName: 'John',
    lastName: 'Doe',
    email: 'john.doe@email.com',
    phone: '+33611223344',
  },
});

const mockUserRepository = {
  findOneById: jest
    .fn()
    .mockImplementationOnce(() => {
      throw new NotFoundException('Record not found');
    })
    .mockImplementation(() => userToUpdate),
  update: jest
    .fn()
    .mockImplementationOnce(() => 'c97b1783-76cf-4840-b298-b90b13c58894')
    .mockImplementationOnce(() => {
      throw new UniqueConstraintException('email already exists');
    })
    .mockImplementationOnce(() => {
      throw new UniqueConstraintException('phone already exists');
    })
    .mockImplementationOnce(() => {
      throw new Error();
    }),
};

describe('update-user.service', () => {
  let updateUserService: UpdateUserService;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        {
          provide: USER_REPOSITORY,
          useValue: mockUserRepository,
        },
        UpdateUserService,
      ],
    }).compile();

    updateUserService = module.get<UpdateUserService>(UpdateUserService);
  });

  it('should be defined', () => {
    expect(updateUserService).toBeDefined();
  });

  describe('execution', () => {
    it('should throw an exception if user is not found', async () => {
      const updateUserCommand = new UpdateUserCommand(
        updateFirstNameUserRequest,
      );
      await expect(
        updateUserService.execute(updateUserCommand),
      ).rejects.toBeInstanceOf(NotFoundException);
    });
    it('should update a user firstName', async () => {
      jest.spyOn(userToUpdate, 'update');
      const updateUserCommand = new UpdateUserCommand(
        updateFirstNameUserRequest,
      );
      const result: AggregateID =
        await updateUserService.execute(updateUserCommand);
      expect(result).toBe('c97b1783-76cf-4840-b298-b90b13c58894');
      expect(userToUpdate.update).toHaveBeenCalledTimes(1);
    });
    it('should throw an exception if Email already exists', async () => {
      const updateUserCommand = new UpdateUserCommand(updateEmailUserRequest);
      await expect(
        updateUserService.execute(updateUserCommand),
      ).rejects.toBeInstanceOf(EmailAlreadyExistsException);
    });
    it('should throw an exception if Phone already exists', async () => {
      const updateUserCommand = new UpdateUserCommand(updatePhoneUserRequest);
      await expect(
        updateUserService.execute(updateUserCommand),
      ).rejects.toBeInstanceOf(PhoneAlreadyExistsException);
    });
    it('should throw an error if something bad happens', async () => {
      const updateUserCommand = new UpdateUserCommand(updatePhoneUserRequest);
      await expect(
        updateUserService.execute(updateUserCommand),
      ).rejects.toBeInstanceOf(Error);
    });
  });
});
