import {
  DatabaseErrorException,
  NotFoundException,
} from '@mobicoop/ddd-library';
import { RpcExceptionCode } from '@mobicoop/ddd-library';
import { DeleteUserGrpcController } from '@modules/user/interface/grpc-controllers/delete-user.grpc.controller';
import { DeleteUserRequestDto } from '@modules/user/interface/grpc-controllers/dtos/delete-user.request.dto';
import { CommandBus } from '@nestjs/cqrs';
import { RpcException } from '@nestjs/microservices';
import { Test, TestingModule } from '@nestjs/testing';

const deleteUserRequest: DeleteUserRequestDto = {
  id: '78153e03-4861-4f58-a705-88526efee53b',
};

const mockCommandBus = {
  execute: jest
    .fn()
    .mockImplementationOnce(() => ({}))
    .mockImplementationOnce(() => {
      throw new NotFoundException();
    })
    .mockImplementationOnce(() => {
      throw new DatabaseErrorException();
    })
    .mockImplementationOnce(() => {
      throw new Error();
    }),
};

describe('Delete User Grpc Controller', () => {
  let deleteUserGrpcController: DeleteUserGrpcController;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        {
          provide: CommandBus,
          useValue: mockCommandBus,
        },
        DeleteUserGrpcController,
      ],
    }).compile();

    deleteUserGrpcController = module.get<DeleteUserGrpcController>(
      DeleteUserGrpcController,
    );
  });

  afterEach(async () => {
    jest.clearAllMocks();
  });

  it('should be defined', () => {
    expect(deleteUserGrpcController).toBeDefined();
  });

  it('should delete a user', async () => {
    jest.spyOn(mockCommandBus, 'execute');
    await deleteUserGrpcController.delete(deleteUserRequest);
    expect(mockCommandBus.execute).toHaveBeenCalledTimes(1);
  });

  it('should throw a dedicated RpcException if user does not exist', async () => {
    jest.spyOn(mockCommandBus, 'execute');
    expect.assertions(3);
    try {
      await deleteUserGrpcController.delete(deleteUserRequest);
    } catch (e: any) {
      expect(e).toBeInstanceOf(RpcException);
      expect(e.error.code).toBe(RpcExceptionCode.NOT_FOUND);
    }
    expect(mockCommandBus.execute).toHaveBeenCalledTimes(1);
  });

  it('should throw a dedicated RpcException if a database error occurs', async () => {
    jest.spyOn(mockCommandBus, 'execute');
    expect.assertions(3);
    try {
      await deleteUserGrpcController.delete(deleteUserRequest);
    } catch (e: any) {
      expect(e).toBeInstanceOf(RpcException);
      expect(e.error.code).toBe(RpcExceptionCode.INTERNAL);
    }
    expect(mockCommandBus.execute).toHaveBeenCalledTimes(1);
  });

  it('should throw a generic RpcException', async () => {
    jest.spyOn(mockCommandBus, 'execute');
    expect.assertions(3);
    try {
      await deleteUserGrpcController.delete(deleteUserRequest);
    } catch (e: any) {
      expect(e).toBeInstanceOf(RpcException);
      expect(e.error.code).toBe(RpcExceptionCode.UNKNOWN);
    }
    expect(mockCommandBus.execute).toHaveBeenCalledTimes(1);
  });
});
