import { NotFoundException } from '@mobicoop/ddd-library';
import { RpcExceptionCode } from '@mobicoop/ddd-library';
import { FindUserByIdGrpcController } from '@modules/user/interface/grpc-controllers/find-user-by-id.grpc.controller';
import { UserMapper } from '@modules/user/user.mapper';
import { QueryBus } from '@nestjs/cqrs';
import { RpcException } from '@nestjs/microservices';
import { Test, TestingModule } from '@nestjs/testing';

const mockQueryBus = {
  execute: jest
    .fn()
    .mockImplementationOnce(() => '200d61a8-d878-4378-a609-c19ea71633d2')
    .mockImplementationOnce(() => {
      throw new NotFoundException();
    })
    .mockImplementationOnce(() => {
      throw new Error();
    }),
};

const mockUserMapper = {
  toResponse: jest.fn().mockImplementationOnce(() => ({
    firstName: 'John',
    lastName: 'Doe',
    email: 'john.doe@email.com',
    phone: '+33611223344',
  })),
};

describe('Find User By Id Grpc Controller', () => {
  let findUserbyIdGrpcController: FindUserByIdGrpcController;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        {
          provide: QueryBus,
          useValue: mockQueryBus,
        },
        {
          provide: UserMapper,
          useValue: mockUserMapper,
        },
        FindUserByIdGrpcController,
      ],
    }).compile();

    findUserbyIdGrpcController = module.get<FindUserByIdGrpcController>(
      FindUserByIdGrpcController,
    );
  });

  afterEach(async () => {
    jest.clearAllMocks();
  });

  it('should be defined', () => {
    expect(findUserbyIdGrpcController).toBeDefined();
  });

  it('should return a user', async () => {
    jest.spyOn(mockQueryBus, 'execute');
    jest.spyOn(mockUserMapper, 'toResponse');
    const response = await findUserbyIdGrpcController.findOnebyId({
      id: '6dcf093c-c7db-4dae-8e9c-c715cebf83c7',
    });
    expect(response.firstName).toBe('John');
    expect(mockQueryBus.execute).toHaveBeenCalledTimes(1);
    expect(mockUserMapper.toResponse).toHaveBeenCalledTimes(1);
  });

  it('should throw a dedicated RpcException if user is not found', async () => {
    jest.spyOn(mockQueryBus, 'execute');
    jest.spyOn(mockUserMapper, 'toResponse');
    expect.assertions(4);
    try {
      await findUserbyIdGrpcController.findOnebyId({
        id: 'ac85f5f4-41cd-4c5d-9aee-0a1acb176fb8',
      });
    } catch (e: any) {
      expect(e).toBeInstanceOf(RpcException);
      expect(e.error.code).toBe(RpcExceptionCode.NOT_FOUND);
    }
    expect(mockQueryBus.execute).toHaveBeenCalledTimes(1);
    expect(mockUserMapper.toResponse).toHaveBeenCalledTimes(0);
  });

  it('should throw a generic RpcException', async () => {
    jest.spyOn(mockQueryBus, 'execute');
    jest.spyOn(mockUserMapper, 'toResponse');
    expect.assertions(4);
    try {
      await findUserbyIdGrpcController.findOnebyId({
        id: '53c8e7ec-ef68-42bc-ba4c-5ef3effa60a6',
      });
    } catch (e: any) {
      expect(e).toBeInstanceOf(RpcException);
      expect(e.error.code).toBe(RpcExceptionCode.UNKNOWN);
    }
    expect(mockQueryBus.execute).toHaveBeenCalledTimes(1);
    expect(mockUserMapper.toResponse).toHaveBeenCalledTimes(0);
  });
});
